/*
 * Copyright (c) 2002-2009 syntevo GmbH. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  o Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  o Neither the name of syntevo GmbH nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.syntevo.q.gui.layout;

import java.awt.*;
import java.util.*;
import java.util.List;
import javax.swing.*;

/**
 * @author Thomas Singer
 */
public final class TableGridLayout extends AbstractLayout {

	// Constants ==============================================================

	private static final SizeProvider MIN_WIDTH_PROVIDER = new SizeProvider() {
		@Override
		public String toString() {
			return "minWidth";
		}

		public int getSize(Component component) {
			return component.getMinimumSize().width;
		}
	};

	private static final SizeProvider MIN_HEIGHT_PROVIDER = new SizeProvider() {
		@Override
		public String toString() {
			return "minHeight";
		}

		public int getSize(Component component) {
			return component.getMinimumSize().height;
		}
	};

	private static final SizeProvider PREF_WIDTH_PROVIDER = new SizeProvider() {
		@Override
		public String toString() {
			return "prefWidth";
		}

		public int getSize(Component component) {
			return component.getPreferredSize().width;
		}
	};

	private static final SizeProvider PREF_HEIGHT_PROVIDER = new SizeProvider() {
		@Override
		public String toString() {
			return "prefHeight";
		}

		public int getSize(Component component) {
			return component.getPreferredSize().height;
		}
	};

	// Static =================================================================

	private static TableGridLayout debugInstance;

	public static void setDebugInstance(TableGridLayout debugInstance) {
		TableGridLayout.debugInstance = debugInstance;
	}

	// Fields =================================================================

	private final List<TableGridColumnOrRow> columns = new ArrayList<TableGridColumnOrRow>();
	private final List<TableGridColumnOrRow> rows = new ArrayList<TableGridColumnOrRow>();
	private final Map<Component, TableGridConstraints> constraints = new HashMap<Component, TableGridConstraints>();
	private final JComponent container;
	private final boolean ignoreInvisibleComponents;

	private int[] cacheColumnsX;
	private int[] cacheRowsY;
	private Dimension cachePreferredLayoutSize;
	private Dimension cacheMinimumLayoutSize;

	// Setup ==================================================================

	TableGridLayout() {
		this.container = null;
		this.ignoreInvisibleComponents = false;
	}

	public TableGridLayout(JComponent container, boolean ignoreInvisibleComponents) {
		if (container == null) {
			throw new IllegalArgumentException("expected not-null container");
		}

		this.container = container;
		this.ignoreInvisibleComponents = ignoreInvisibleComponents;
	}

	// Implemented ============================================================

	@Override
	protected void doAddLayoutComponent(Component component, Object constraints) {
		if (component == null) {
			throw new IllegalArgumentException("expected not-null component");
		}
		if (constraints == null) {
			throw new IllegalArgumentException("expected not-null constraints");
		}

		if (container != null) {
			if (container == component) {
				throw new IllegalArgumentException("cannot add container to itself");
			}
			final Container parent = component.getParent();
			if (parent != null && parent != container) {
				throw new IllegalStateException("invalid component parent");
			}
		}

		final TableGridConstraints higConstraints = (TableGridConstraints)constraints;

		if (higConstraints.getColumn() + higConstraints.getWidth() > columns.size()) {
			throw new IllegalStateException("Not enough columns defined. Available: " + columns.size() + ", required: " + (higConstraints.getColumn() + higConstraints.getWidth()));
		}
		if (higConstraints.getRow() + higConstraints.getHeight() > rows.size()) {
			throw new IllegalStateException("Not enough rows defined. Available: " + rows.size() + ", required: " + (higConstraints.getRow() + higConstraints.getHeight()));
		}
		if (this.constraints.containsKey(component)) {
			throw new IllegalStateException("Component already was added");
		}

		// if comp. occupies one column (row), we insert it to list for this column (row)
		if (higConstraints.getWidth() == 1) {
			getColumn(higConstraints.getColumn()).addComponent(component);
		}
		if (higConstraints.getHeight() == 1) {
			getRow(higConstraints.getRow()).addComponent(component);
		}
		this.constraints.put(component, higConstraints);
	}

	@Override
	protected void doRemoveLayoutComponent(Component component) {
		if (component == null) {
			throw new IllegalArgumentException("expected not-null component");
		}

		if (container != null) {
			final Container parent = component.getParent();
			if (parent != null && parent != container) {
				throw new IllegalStateException("invalid component parent");
			}
		}

		final TableGridConstraints constraints = this.constraints.remove(component);
		if (constraints == null) {
			return;
		}

		getColumn(constraints.getColumn()).removeComponent(component);
		getRow(constraints.getRow()).removeComponent(component);
	}

	@Override
	protected void doLayoutContainer(Dimension size, Insets insets) {
		if (isDebugInstance()) {
			System.out.println("layoutContainer");
		}

		assertComponents();

		int sizeWidth = size.width;
		int sizeHeight = size.height;
		sizeWidth -= insets.left + insets.right;
		sizeHeight -= insets.top + insets.bottom;

		final int[] xCoordinates = getColumnsX(sizeWidth, insets);
		final int[] yCoordinates = getRowsY(sizeHeight, insets);

		for (final Component component : getComponents()) {
			alignComponentInCell(component, xCoordinates, yCoordinates);
		}
	}

	@Override
	protected Dimension getPreferredLayoutSize(Insets insets) {
		if (isDebugInstance()) {
			System.out.println("getPreferredLayoutSize");
		}

		assertComponents();

		if (cachePreferredLayoutSize == null) {
			calculateLayoutSizes(insets);

			if (isDebugInstance()) {
				System.out.println(cachePreferredLayoutSize);
			}
		}
		return new Dimension(cachePreferredLayoutSize);
	}

	@Override
	protected Dimension getMinimumLayoutSize(Insets insets) {
		if (isDebugInstance()) {
			System.out.println("getMinimumLayoutSize");
		}

		assertComponents();

		if (cacheMinimumLayoutSize == null) {
			calculateLayoutSizes(insets);

			if (isDebugInstance()) {
				System.out.println(cacheMinimumLayoutSize);
			}
		}
		return new Dimension(cacheMinimumLayoutSize);
	}

	/**
	 * Invalidates the layout, indicating that if the layout manager
	 * has cached information it should be discarded.
	 */
	public void invalidateLayout(Container target) {
		if (container != null && container != target) {
			throw new IllegalStateException("unexpected target specified");
		}

		assertComponents();

		cacheMinimumLayoutSize = null;
		cachePreferredLayoutSize = null;
		cacheColumnsX = null;
		cacheRowsY = null;
	}

	// Accessing ==============================================================

	public int getLastColumnIndex() {
		return columns.size() - 1;
	}

	public int getLastRowIndex() {
		return rows.size() - 1;
	}

	public int addColumn(int width, TableGridStretchValue stretchValue, TableGridGroup columnGroup) {
		if (stretchValue == null) {
			throw new NullPointerException("stretchValue must not be null");
		}

		columns.add(new TableGridColumnOrRow(width, stretchValue, columnGroup));
		return getLastColumnIndex();
	}

	public int addRow(int height, TableGridStretchValue stretchValue, TableGridGroup rowGroup) {
		if (stretchValue == null) {
			throw new NullPointerException("stretchValue must not be null");
		}

		rows.add(new TableGridColumnOrRow(height, stretchValue, rowGroup));
		return getLastRowIndex();
	}

	// Utils ==================================================================

	private void assertComponents() {
		if (container == null) {
			return;
		}

		for (final Component component : constraints.keySet()) {
			if (component.getParent() != container) {
				throw new IllegalStateException("component " + component + " has invalid parent (expected: " + container + ")");
			}
		}
	}

	private TableGridColumnOrRow getColumn(int index) {
		return columns.get(index);
	}

	private TableGridColumnOrRow getRow(int index) {
		return rows.get(index);
	}

	private TableGridConstraints getConstraints(Component comp) {
		return constraints.get(comp);
	}

	private void distributeSizeDifference(int desiredSize, int[] sizes, int[] minSizes, List<TableGridColumnOrRow> columnOrRows) {
		final int weightSum = calculateStretchOnResizeSum(columnOrRows);

		if (weightSum == 0) {
			return;
		}

		int preferredSize = 0;
		for (int size : sizes) {
			preferredSize += size;
		}

		final double unit = ((double)(desiredSize - preferredSize)) / (double)weightSum;

		int i = 0;
		for (final TableGridColumnOrRow columnOrRow : columnOrRows) {
			final int weight = (columnOrRow.getStretchValue() == TableGridStretchValue.STRETCH_ON_RESIZE)
					? 1
					: 0;
			final int newLength = sizes[i] + (int)(unit * weight);
			sizes[i] = (newLength > minSizes[i])
					? newLength
					: minSizes[i];
			i++;
		}

		// current limitation: if multiple STRETCH_ON_RESIZE columns exists, e.g. 2, the unit can become
		// a broken value, e.g. 2.5, which causes the newLength value to always be rounded down
		// Example: desiredSize - preferredSize = 5
		//          2 columns with STRETCH_ON_RESIZE
		//          => unit = 2.5
		//          => for both stretching columns only 2 pixels are added
		//          => one pixel remains undistributed!
	}

	private int[] calcSizes(List<TableGridColumnOrRow> columnOrRows, SizeProvider sizeProvider) {
		final int[] sizes = new int[columnOrRows.size()];
		int i = 0;
		for (final TableGridColumnOrRow columnOrRow : columnOrRows) {
			if (columnOrRow.getSize() > 0) {
				if (columnOrRow.getStretchValue() == TableGridStretchValue.NO_STRETCH) {
					sizes[i] = columnOrRow.getSize();
				}
				else {
					final int sizeFromComponents = getSizeFromComponents(columnOrRow, sizeProvider);
					sizes[i] = Math.max(columnOrRow.getSize(), sizeFromComponents);
				}
			}
			else {
				sizes[i] = getSizeFromComponents(columnOrRow, sizeProvider);
			}
			i++;
		}
		return sizes;
	}

	private int getSizeFromComponents(TableGridColumnOrRow columnOrRow, SizeProvider sizeProvider) {
		int maxSize = 0;
		for (final Component component : columnOrRow.getGroupingComponents()) {
			if (ignoreInvisibleComponents && !component.isVisible()) {
				continue;
			}

			final int size = sizeProvider.getSize(component);
			maxSize = Math.max(size, maxSize);
		}
		return maxSize;
	}

	/**
	 * returns array of x-coordinates of columns. First coordinate is stored in x[1]
	 * Reference to this array is cached, so data should not be modified.
	 */
	private int[] getColumnsX(int targetWidth, Insets insets) {
		if (cacheColumnsX != null) {
			return cacheColumnsX;
		}

		final int[] preferredColumnWidths = calcSizes(columns, PREF_WIDTH_PROVIDER);
		final int[] minColumnWidths = calcSizes(columns, MIN_WIDTH_PROVIDER);
		ensureMinSizes(preferredColumnWidths, minColumnWidths);

		for (final Component component : getComponents()) {
			final TableGridConstraints constraints = getConstraints(component);
			final int columnCount = constraints.getWidth();
			if (columnCount > 1) {
				distributeSizes(component, PREF_WIDTH_PROVIDER, columns, constraints.getColumn(), columnCount, preferredColumnWidths);
			}
		}

		distributeSizeDifference(targetWidth, preferredColumnWidths, minColumnWidths, columns);
		cacheColumnsX = getCoordinates(preferredColumnWidths, (insets != null)
				? insets.left
				: 0);
		return cacheColumnsX;
	}

	/**
	 * returns array of y-coordinates of rows. First coordinate is stored in y[1].
	 * Reference to this array is cached, so data should not be modified.
	 */
	private int[] getRowsY(int targetHeight, Insets insets) {
		if (cacheRowsY != null) {
			return cacheRowsY;
		}

		final int[] preferredRowHeights = calcSizes(rows, PREF_HEIGHT_PROVIDER);
		final int[] minRowHeights = calcSizes(rows, MIN_HEIGHT_PROVIDER);
		ensureMinSizes(preferredRowHeights, minRowHeights);

		for (final Component component : getComponents()) {
			final TableGridConstraints constraints = getConstraints(component);
			final int rowCount = constraints.getHeight();
			if (rowCount > 1) {
				distributeSizes(component, PREF_HEIGHT_PROVIDER, rows, constraints.getRow(), rowCount, preferredRowHeights);
			}
		}

		distributeSizeDifference(targetHeight, preferredRowHeights, minRowHeights, rows);
		cacheRowsY = getCoordinates(preferredRowHeights, (insets != null)
				? insets.top
				: 0);
		return cacheRowsY;
	}

	private void distributeSizes(Component component, SizeProvider sizeProvider, List<TableGridColumnOrRow> columnOrRows,
	                             int startIndex, int length, int[] sizes) {
		int stretchOnResizeSum = 0;
		int stretchIfRequiredSum = 0;
		int size = 0;
		for (int i = 0, index = startIndex; i < length; i++, index++) {
			final TableGridColumnOrRow columnOrRow = columnOrRows.get(index);
			final TableGridStretchValue stretchValue = columnOrRow.getStretchValue();
			if (stretchValue == TableGridStretchValue.STRETCH_ON_RESIZE) {
				stretchOnResizeSum++;
			}
			else if (stretchValue == TableGridStretchValue.STRETCH_IF_REQUIRED) {
				stretchIfRequiredSum++;
			}
			size += sizes[index];
		}

		final int sizeDiff = sizeProvider.getSize(component) - size;
		if (sizeDiff <= 0) {
			return;
		}
		if (stretchOnResizeSum == 0 && stretchIfRequiredSum == 0) {
			return;
		}

		final int stretchSum = (stretchOnResizeSum > 0)
				? stretchOnResizeSum
				: stretchIfRequiredSum;
		int sizeToDistribute = sizeDiff;
		int weightsToDistribute = stretchSum;

		for (int i = 0, index = startIndex; i < length; i++, index++) {
			final TableGridColumnOrRow columnOrRow = columnOrRows.get(index);
			final TableGridStretchValue stretchValue = columnOrRow.getStretchValue();
			if (stretchValue == TableGridStretchValue.NO_STRETCH) {
				continue;
			}

			if (stretchValue == TableGridStretchValue.STRETCH_IF_REQUIRED && stretchOnResizeSum > 0) {
				continue;
			}

			int delta = sizeDiff / stretchSum;

			sizeToDistribute -= delta;
			weightsToDistribute--;
			if (weightsToDistribute == 0) {
				// distribute the rest
				delta += sizeToDistribute;
			}

			sizes[index] += delta;
		}
	}

	private void alignComponentInCell(Component component, int[] x, int[] y) {
		final TableGridConstraints constraints = getConstraints(component);

		// first we center component into cell
		final Dimension prefSize = component.getPreferredSize();
		int width = prefSize.width;
		int height = prefSize.height;
		final int cellWidth;
		if (constraints.getWidth() < 0) {
			width = -constraints.getWidth();
			cellWidth = x[constraints.getColumn() + 1] - x[constraints.getColumn()];
		}
		else {
			cellWidth = x[constraints.getColumn() + constraints.getWidth()] - x[constraints.getColumn()];
		}

		final int cellHeight;
		if (constraints.getHeight() < 0) {
			height = -constraints.getHeight();
			cellHeight = y[constraints.getRow() + 1] - y[constraints.getRow()];
		}
		else {
			cellHeight = y[constraints.getRow() + constraints.getHeight()] - y[constraints.getRow()];
		}

		// I had intend to ensure that maximumSize is respected, but Swing components returns stupid maximumSize
		final float dWidth = (cellWidth - width) / 2.0f;
		final float dHeight = (cellHeight - height) / 2.0f;
		float componentX = x[constraints.getColumn()] + dWidth;
		float componentY = y[constraints.getRow()] + dHeight;

		// now anchor to cell borders
		boolean xSized = false;	// first move, then change width (when opposite border)
		boolean ySized = false;
		if (constraints.isStickLeft()) {
			componentX = x[constraints.getColumn()];
			xSized = true;
		}
		if (constraints.isStickTop()) {
			componentY = y[constraints.getRow()];
			ySized = true;
		}
		if (constraints.isStickRight()) {
			if (xSized) {
				width = cellWidth;
			}
			else {
				componentX += dWidth;
			}
		}
		if (constraints.isStickBottom()) {
			if (ySized) {
				height = cellHeight;
			}
			else {
				componentY += dHeight;
			}
		}

		component.setBounds((int)componentX, (int)componentY, width, height);
	}

	private Collection<Component> getComponents() {
		if (ignoreInvisibleComponents) {
			final Collection<Component> components = new ArrayList<Component>();
			for (final Component component : constraints.keySet()) {
				if (component.isVisible()) {
					components.add(component);
				}
			}
			return components;
		}
		return constraints.keySet();
	}

	private void calculateLayoutSizes(Insets insets) {
		final int[] prefColumnWidths = calcSizes(columns, PREF_WIDTH_PROVIDER);
		final int[] prefRowHeights = calcSizes(rows, PREF_HEIGHT_PROVIDER);
		final int[] minColumnWidths = calcSizes(columns, MIN_WIDTH_PROVIDER);
		final int[] minRowHeights = calcSizes(rows, MIN_HEIGHT_PROVIDER);

		for (final Component component : getComponents()) {
			final TableGridConstraints constraints = getConstraints(component);
			final int columnCount = constraints.getWidth();
			if (columnCount > 1) {
				final int startColumn = constraints.getColumn();
				distributeSizes(component, PREF_WIDTH_PROVIDER, columns, startColumn, columnCount, prefColumnWidths);
				distributeSizes(component, MIN_WIDTH_PROVIDER, columns, startColumn, columnCount, minColumnWidths);
			}
			final int rowCount = constraints.getHeight();
			if (rowCount > 1) {
				final int startRow = constraints.getRow();
				distributeSizes(component, PREF_HEIGHT_PROVIDER, rows, startRow, rowCount, prefRowHeights);
				distributeSizes(component, MIN_HEIGHT_PROVIDER, rows, startRow, rowCount, minRowHeights);
			}
		}

		cachePreferredLayoutSize = new Dimension(sumArrayElements(prefColumnWidths) + insets.left + insets.right,
		                                         sumArrayElements(prefRowHeights) + insets.top + insets.bottom);
		cacheMinimumLayoutSize = new Dimension(sumArrayElements(minColumnWidths) + insets.left + insets.right,
		                                       sumArrayElements(minRowHeights) + insets.top + insets.bottom);
	}

	private boolean isDebugInstance() {
		return debugInstance == this;
	}

	private static int calculateStretchOnResizeSum(List<TableGridColumnOrRow> columnOrRows) {
		int weightsSum = 0;
		for (final TableGridColumnOrRow column : columnOrRows) {
			if (column.getStretchValue() == TableGridStretchValue.STRETCH_ON_RESIZE) {
				weightsSum++;
			}
		}
		return weightsSum;
	}

	private static void ensureMinSizes(int[] preferredSizes, int[] minSizes) {
		for (int i = preferredSizes.length; i-- > 0;) {
			preferredSizes[i] = Math.max(preferredSizes[i], minSizes[i]);
		}
	}

	private static int[] getCoordinates(int[] sizes, int startOffset) {
		final int[] x = new int[sizes.length + 1];
		x[0] = startOffset;

		for (int i = 1; i < x.length; i++) {
			x[i] = x[i - 1] + sizes[i - 1];
		}
		return x;
	}

	private static int sumArrayElements(int[] array) {
		int sum = 0;
		for (int i = array.length; i-- > 0;) {
			sum += array[i];
		}
		return sum;
	}

	// Inner Classes ==========================================================

	private interface SizeProvider {
		int getSize(Component component);
	}
}
